// Defaults script; update some default values
//
// Copyright (c) 2011 aizo ag, Switzerland
// Author: Johannes Winkelmann, johannes.winkelmann@aizo.com
//         Christian Hitz, christian.hitz@aizo.com
//
// Compile: g++ -Wall -o pp_set_defaults set_defaults.cpp -ldsm-api-v2

#include <iostream>
#include <string>
#include <iomanip>
#include <sstream>

#include <cstdlib>

#include <getopt.h>
#include <stdint.h>

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>

struct ParamSet {
  int16_t bank;
  int16_t offset;
  int16_t value;
  int16_t minVersion;
};
#define PARAM_END { -1, -1, -1, -1 }

// ----
// Configuration values

ParamSet kChipParameter[] = {
  { 0x4, 0x0c, 0x76, 0x0302 },
  { 0x4, 0x14, 0x14, 0x0302 },
  { 0x4, 0x09, 0x49, 0x0301 },
  { 0x4, 0x0a, 0x91, 0x0301 },
  { 0x4, 0x1d, 0x20, 0x0314 },
  PARAM_END
};

ParamSet kAreaParameter[] = {
  { 0x80, 0x86, 0x56, 0x0314 },
  { 0x80, 0x87, 0x56, 0x0314 },
  { 0x80, 0x88, 0x56, 0x0314 },
  { 0x80, 0x89, 0x56, 0x0314 },
  PARAM_END
};

// ----


const uint8_t kErrPing = 1;
const uint8_t kErrChip = 4;

const uint8_t kRetryCount = 3;

const std::string TOP_PREFIX = "> ";
const std::string INNER_PREFIX = ">> ";

std::string dsid2string(dsid_t& dsid)
{
  std::ostringstream str;
  for (int i = 0; i < 12; ++i) {
    str << std::setfill('0') << std::setw(2) << std::hex << (int)dsid.id[i];
    if (i < 11) {
      str << ":";
    }
  }

  return str.str();
}

uint8_t configurePing(void* handle)
{
  std::cout << TOP_PREFIX << "Configuring ping settings..." << std::endl;
  uint8_t retval = 0;
  uint8_t deviceCount = 0;
  uint8_t dsmCount = 0;

  dsid_t device_list[63];
  int ret = DsmApiGetBusMembers(handle, device_list, 63);
  if (ret < 0) {
    return kErrPing;
  }

  dsmCount = ret;
  for (int i = 0; i < dsmCount; ++i) {
    if (!DsmApiIsdSM(device_list[i])) {
      continue;
    }
    uint8_t flags = 0;
    ret = dSMProperties_get_flags(handle, device_list[i], &flags);
    if (ret != 0) {
      std::cout << INNER_PREFIX << "Getting dsm flags failed" << std::endl;
      retval = kErrPing;
    } else {
      flags |= 6;

      ret = dSMProperties_set_flags(handle, device_list[i], flags);
      if (ret != 0) {
        std::cout << INNER_PREFIX << "Setting dsm flags failed" << std::endl;
        retval = kErrPing;
      }
    }

    ret = dSMProperties_set_ping_intervals(handle, device_list[i], 20000, 60000);
    if (ret != 0) {
      std::cout << INNER_PREFIX << "Setting ping intervals failed" << std::endl;
      retval = kErrPing;
    }
    ++deviceCount;
  }

  std::cout << TOP_PREFIX << "done (" << (int)deviceCount << " dSMs)" << std::endl;
  return retval;
}

uint8_t configureChip(void* handle, uint8_t configClass, uint8_t configOffset, uint8_t value, uint8_t minVersion, uint8_t group=0)
{
  std::cout << TOP_PREFIX << "Configuring chip settings..."
            << "(Bank "
            << "0x" << std::hex << std::setfill('0') << std::setw(2)
            << (int)configClass << ", "
            << "0x" << std::hex << std::setfill('0') << std::setw(2)
            << (int)configOffset << ", "
            << "0x" << std::hex << std::setfill('0') << std::setw(2)
            << (int)value
            << ")"
            << std::endl;
  uint8_t retval = 0;
  uint8_t tryCount = 0;
  uint8_t deviceCount = 0;
  uint8_t dsmCount = 0;

  dsid_t device_list[63];
  int ret = DsmApiGetBusMembers(handle, device_list, 63);
  if (ret < 0) {
    return kErrChip;
  }

  dsmCount = ret;
  for (int i = 0; i < dsmCount; ++i) {
    if (!DsmApiIsdSM(device_list[i])) {
      continue;
    }	

    uint16_t numberOfDevices;
    ret = ZoneDeviceCount_only_active(handle, device_list[i], 0, &numberOfDevices);
    if (ret != 0) {
      std::cout << INNER_PREFIX << "Getting device count failed" << std::endl;
      retval = kErrChip;
      continue;
    }

    for (int j = 0; j < numberOfDevices; ++j) {
      uint16_t deviceID;
      uint16_t vendorID;
      uint16_t functionID;
      uint16_t version;
      uint32_t serialNumber;
      dsid_t devdsid;

      // - read out the device ID
      tryCount = 0;
      do {
        ret = DeviceInfo_by_index_only_active(handle, device_list[i], 0, j,
                                              &deviceID, &vendorID, NULL,
                                              &functionID, &version, NULL, NULL, NULL,
                                              NULL, NULL, NULL, NULL,
                                              &serialNumber);
        ++tryCount;
      } while (ret != 0 && tryCount < kRetryCount);

      if ((group != 0) && ((functionID >> 12) != group)) {
        continue;
      }

      if (version < minVersion) {
        continue;
      }

      DsmApiExpandDeviceDSID(vendorID, serialNumber, &devdsid);

      // - try setting the parameter
      tryCount = 0;
      do {
        ret = DeviceConfig_set(handle, device_list[i], deviceID, configClass, configOffset, value);
        ++tryCount;
      } while (ret != 0 && tryCount < kRetryCount);

      if (ret != 0) {
        std::cout << "\n" << INNER_PREFIX << "Setting device config failed: "
                  << dsid2string(devdsid) << std::endl;
        retval = kErrChip;
        continue;
      }

      // - read the parameter
      uint8_t readValue;
      uint8_t timeout = 10;
      tryCount = 0;
      do {
        ret = DeviceConfig_get_sync_8(handle, device_list[i], deviceID,
                                      configClass, configOffset, timeout, &readValue);
        ++tryCount;
      } while (ret != 0 && tryCount < kRetryCount);

      if (ret != 0) {
        std::cout << "\n" << INNER_PREFIX << "Getting device config failed: "
                  << dsid2string(devdsid) << std::endl;
        retval = kErrChip;
        continue;
      }

      // - compare read out value to intended one
      if (value != readValue) {
        std::cout << "\n" << INNER_PREFIX << "Failed to set param on device: "
                  << dsid2string(devdsid) << std::endl;
        retval = kErrChip;
      }

      std::cout << ".";
      std::cout.flush();

      ++deviceCount;
      // verify
    }
  }

  std::cout << "\n" << TOP_PREFIX << "done ("
            << std::dec << (int)deviceCount << " devices)" << std::endl;
  return retval;
}

void usage(const char* appname)
{
  std::cout << "Usage: " << basename(appname) << " [options]" << std::endl;
  std::cout << "    where options can be: " << std::endl;
  std::cout << "    -c configure chip settings" << std::endl;
  std::cout << "    -p configure ping settings" << std::endl;
  std::cout << "    -a fix broken area settings" << std::endl;
}

int main(int argc, char** argv)
{
  bool doPing = false;
  bool doChip = false;
  bool doArea = false;
  uint8_t retval = 0;


  int opt;
  while ((opt = getopt(argc, argv, "cpTah")) != -1) {
    switch (opt) {
      case 'c':
        doChip = true;
        break;
      case 'p':
        doPing = true;
        break;
      case 'a':
        doArea = true;
        break;
      case 'h':
        usage(argv[0]);
        exit(0);
      case 'T':
        // Test / debug switch
        dsid_t dsid;
        for (int i = 0; i < 12; ++i) {
          dsid.id[i] = i;
        }
        std::cout << dsid2string(dsid) << std::endl;
        exit(0);
      default: /* '?' */
        usage(argv[0]);
        exit(1);
    }
  }

  if (!doChip && !doPing && !doArea) {
    usage(argv[0]);
    exit(1);
  }


  std::string connection = "tcp://localhost:8442";
  if (optind < argc) {
    connection = argv[optind];
  }

  void* dsmApiHandle = DsmApiInitialize();
  int result = DsmApiOpen(dsmApiHandle, connection.c_str(), 0);
  if (result < 0) {
    std::cout << "Couldn't open dsmapi connection "  << connection << std::endl;
    return -1;
  }

  if (doPing) {
    retval |= configurePing(dsmApiHandle);
  }
  if (doChip) {
    const ParamSet* p = kChipParameter;
    while (p->bank != -1) {
      retval |= configureChip(dsmApiHandle, p->bank, p->offset, p->value, p->minVersion);
      ++p;
    }
  }
  if (doArea) {
    const ParamSet* p = kAreaParameter;
    while (p->bank != -1) {
      static const uint8_t groups[] = { 1, 2, 8 };
      for (int i = 0; i < sizeof(groups); ++i) {
        retval |= configureChip(dsmApiHandle, p->bank, p->offset, p->value, p->minVersion, groups[i]);
      }
      ++p;
    }
  }

  DsmApiCleanup(dsmApiHandle);
  DsmApiClose(dsmApiHandle);

  return retval;
}
